# monkeypatch ZPublisher's call_object so that we can make sure we
# have a trace in each thread before objects are published

# patching approach taken from PDBDebugMode, but I haven't had a
# chance to make this actually coexist with PDBDebugMode yet...

from mr.freeze.freeze import set_thread_trace_when_object_called

from ZPublisher import Publish
from ZPublisher.Publish import missing_name
from ZPublisher.Publish import dont_publish_class
from ZPublisher.Publish import mapply
from Globals import DevelopmentMode

try:
    from Products.PlacelessTranslationService import PatchStringIO
except ImportError:
    from ZPublisher.Publish import publish as real_publish
    USE_PTS = False
else:
    USE_PTS = True
    from ZPublisher.Publish import old_publish as real_publish

real_call_object = real_publish.func_defaults[1]
def freeze_call(object, args, request):
    return set_thread_trace_when_object_called(object, real_call_object, args, request)

def pdb_publish(request, module_name, after_list, debug=0,
                call_object=freeze_call,
                missing_name=missing_name,
                dont_publish_class=dont_publish_class,
                mapply=mapply, ):
    """Hook the publish function to override the function used to call
    the result of the request traversal."""
    return real_publish(request, module_name, after_list, debug=0,
                call_object=call_object,
                missing_name=missing_name,
                dont_publish_class=dont_publish_class,
                mapply=mapply, )

if DevelopmentMode:

    # call set_thread_trace_when_object_called when the final call for a request is done
    if USE_PTS:
        Publish.zpublisher_publish = Publish.old_publish
        Publish.old_publish = pdb_publish
    else:
        Publish.old_publish = Publish.publish
        Publish.publish = pdb_publish



import os
import re
import signal
import sys
import logging
import string
logger = logging.getLogger('mr.freeze')

from App.config import getConfiguration
from mr.freeze.freeze import set_freeze

def initialize(context):
    import patches
    logger.info('Patched ZPublisher.Publish.call_object')
    signal.signal(signal.SIGUSR1, handleSignal)
    logger.info('Installed USR1 signal handler')

def getFreezeFileName():
    cfg = getConfiguration()
    fname, ext = os.path.splitext(cfg.pid_filename)
    return fname + '.freeze'

def readCommands():
    freeze_filename = getFreezeFileName()
    try:
        freeze_file = open(freeze_filename, 'r')
        cmds = freeze_file.readlines()
        freeze_file.close()
        return map(string.strip, cmds)
    except IOError:
        logger.warn('Unable to read freeze file %s' % freeze_filename)
        return None

freeze_re = re.compile(r'^freeze\s+(.+)\s+(\d+)$')
def handleSignal(signal, stack):
    for cmd in readCommands():
        freeze_match = freeze_re.match(cmd)
        if cmd == 'pony':
            showPony()
        elif cmd == 'freeze':
            logger.info('Preparing to freeze')
            set_freeze()
        elif freeze_match:
            filename, lineno = freeze_match.group(1, 2)
            filename = os.path.realpath(filename)
            lineno = int(lineno)
            logger.info('Preparing to freeze at %s line %s' % (filename, lineno))
            set_freeze(filename=filename, lineno=lineno)
        elif cmd == 'stack' or cmd is None: # default
            showStacks()
        else:
            logger.warn('Unrecognized command: %s' % cmd)

### Commands ###

from z3c.deadlockdebugger.threads import dump_threads
def showStacks():
    print dump_threads()
    sys.stdout.flush()

def showPony():
    print """
            .,,.
         ,;;*;;;;,
        .-'``;-');;.
       /'  .-.  /*;;
     .'    \d    \;;               .;;;,
    / o      `    \;    ,__.     ,;*;;;*;,
    \__, _.__,'   \_.-') __)--.;;;;;*;;;;,
     `""`;;;\       /-')_) __)  `\' ';;;;;;
        ;*;;;        -') `)_)  |\ |  ;;;;*;
        ;;;;|        `---`    O | | ;;*;;;
        *;*;\|                 O  / ;;;;;*
       ;;;;;/|    .-------\      / ;*;;;;;
      ;;;*;/ \    |        '.   (`. ;;;*;;;
      ;;;;;'. ;   |          )   \ | ;;;;;;
      ,;*;;;;\/   |.        /   /` | ';;;*;
       ;;;;;;/    |/       /   /__/   ';;;
       '*jgs/     |       /    |      ;*;
            `""""`        `""""`     ;'
"""
    sys.stdout.flush()



import sys
import pdb
from bdb import BdbQuit
from thread import get_ident

# global counter for which freeze we're on
freeze_count = 0
# global flag for whether the freeze has been hit
freeze_pending = False
# file and line to stop on
freeze_file = None
freeze_line = None

# per-thread counter for which freeze has had a trace set
trace_up_to_date = {}

def set_freeze(filename=None, lineno=None):
    global freeze_pending, freeze_count, freeze_file, freeze_line
    freeze_pending = True
    freeze_count += 1
    freeze_file = filename
    freeze_line = lineno

def is_trace_up_to_date():
    """
    Returns true if the current thread already has a trace set for
    the most recent freeze.
    """
    return trace_up_to_date.get(get_ident(), 0) == freeze_count

class Fdb(pdb.Pdb):
    """
    Subclass the standard debugger to:
     * only set and dispatch trace if the thread's freeze_pending flag is on
     * allow initially set a breakpoint rather than startting stepping
     * clear the freeze_pending flag *for all threads* once the breakpoint is hit
    """
    
    # per-instance flag for keeping track of whether we've successfully hit the freezepoint
    freeze_active = False
    
    def set_trace(self, frame=None, filename=None, lineno=None):
        """
        Sets a trace if this thread doesn't have one yet and there is a freeze pending.
        """
        if not freeze_pending or is_trace_up_to_date():
            return
        
        if frame is None:
            frame = sys._getframe().f_back
        self.reset()
        while frame:
            # not sure why our callers need a trace function set when we're
            # setting a global one...this was causing problems when continuing
            # back into those callers (publish loop) when a breakpoint was still set
            #frame.f_trace = self.trace_dispatch
            self.botframe = frame
            frame = frame.f_back
        
        if filename is not None and lineno is not None:
            self.set_break(filename, lineno, temporary=1)
            self.set_continue()
        else:
            self.clear_freeze()
            self.set_step()
        sys.settrace(self.trace_dispatch)
        
        trace_up_to_date[get_ident()] = freeze_count

    def trace_dispatch(self, frame, event, arg):
        """
        Stop tracing if some other thread handled the freeze.
        """
        if not freeze_pending or not is_trace_up_to_date():
            if not self.freeze_active:
                self.set_quit()
        return pdb.Pdb.trace_dispatch(self, frame, event, arg)
    
    def clear_break(self, filename, lineno):
        self.clear_freeze()
        return pdb.Pdb.clear_break(self, filename, lineno)
    
    def clear_freeze(self):
        global freeze_pending
        freeze_pending = False
        self.freeze_active = True

def set_thread_trace_when_object_called(object, func, args, request):
    if freeze_pending and trace_up_to_date.get(get_ident(), 0) < freeze_count:
        if freeze_file is not None and freeze_line is not None:
            Fdb().set_trace(filename=freeze_file, lineno=freeze_line)
        else:
            global freeze_pending
            freeze_pending = False
            return pdb.runcall(object, *args)
    try:
        return func(object, args, request)
    except BdbQuit:
        return 'Request cancelled by quitting pdb.  Try the continue (c) command instead if you want the page to finish rendering.'



from setuptools import setup, find_packages
import os

version = "1.0a3"

setup(name="mr.freeze",
      version=version,
      description="Trigger a pdb in a running zope via a USR1 signal",
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Framework :: Zope2",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords="zope debugging pdb",
      author="David Glick",
      author_email="davidglick@onenw.org",
      url="http://en.wikipedia.org/wiki/Mr._Freeze",
      license="BSD",
      packages=find_packages(exclude=["ez_setup"]),
      namespace_packages=["mr"],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          "setuptools",
          "z3c.deadlockdebugger"
      ],
      )
